﻿Imports Autodesk.AutoCAD.DatabaseServices
Imports Autodesk.AutoCAD.Runtime
Imports Autodesk.AutoCAD.Geometry
Imports Autodesk.AutoCAD.ApplicationServices
Imports Autodesk.AutoCAD.EditorInput
Imports Autodesk.AutoCAD.Colors

Public Class Class1

    <CommandMethod("JoinTheDots")> _
    Public Sub JoinTheDots()
        'Setup
        Dim ed As Editor = Application.DocumentManager.MdiActiveDocument.Editor

        'Declare our filter entries this way
        Dim values() As TypedValue = {
            New TypedValue(DxfCode.Start, "POINT")
            }

        'Create the filter using our values
        Dim sfilter As New SelectionFilter(values)

        'Set the selection options
        Dim SelOpts As New PromptSelectionOptions()
        SelOpts.MessageForAdding = "Select points to find the shortest path through"
        SelOpts.AllowDuplicates = True

        'Make the selection
        Dim res As PromptSelectionResult = ed.GetSelection(SelOpts, sfilter)

        'If the user did something other than make a selection (like pressing ESC), then abort
        If Not res.Status = PromptStatus.OK Then Return

        'Create a selection set based on the user's selection
        Dim ss As Autodesk.AutoCAD.EditorInput.SelectionSet = res.Value

        'Create an array of object IDs from the selection set, so that we can loop through them easily
        Dim idarray As ObjectId() = ss.GetObjectIds()



        'Create a link to the active document's database
        Dim db As Database = Application.DocumentManager.MdiActiveDocument.Database
        'Create a transaction manager object
        Dim tm As Autodesk.AutoCAD.DatabaseServices.TransactionManager = db.TransactionManager
        'Start a transaction
        Dim myT As Transaction = tm.StartTransaction()



        'Create a coord array, the same size as the idarray
        Dim points(idarray.GetUpperBound(0)) As DBPoint
        'Populate the coord array with point2d objects representing the location of the DBPoint objects
        Dim n As Long
        For n = 0 To idarray.GetUpperBound(0)
            points(n) = tm.GetObject(idarray(n), OpenMode.ForRead, True)
        Next


        'Find the shortest line formed by two coordinates, create a polyline representing this line, and
        'remove the two coordinates from the coords array
        Dim pl As Polyline = GetShortestLine(points)


        Dim TimeToExit As Boolean = False
        Do
            TimeToExit = AppendPoint(pl, points)
        Loop Until TimeToExit


        'Add the polyline to modelspace
        AddToModelSpace(db, myT, pl)



        'Commit the transaction
        myT.Commit()
        'Dispose of our transaction
        myT.Dispose()


    End Sub

    Private Function AppendPoint(ByRef pl As Polyline, ByRef points() As DBPoint) As Boolean
        Dim startPoint As Point2d = pl.StartPoint.Convert2d(New Plane)
        Dim endPoint As Point2d = pl.EndPoint.Convert2d(New Plane)
        Dim nearestPoint As String = ""
        Dim shortestDist As Double = 1.0E+300 'A very very big number!
        Dim remPoint As DBPoint

        Dim n As Long
        For n = 0 To points.GetUpperBound(0)

            Dim targetPoint As Point2d
            targetPoint = points(n).Position.Convert2d(New Plane)

            Dim tmpDist As Double
            tmpDist = startPoint.GetDistanceTo(targetPoint)
            If tmpDist < shortestDist Then
                shortestDist = tmpDist
                nearestPoint = "startpoint"
                remPoint = points(n)
            End If

            tmpDist = endPoint.GetDistanceTo(targetPoint)
            If tmpDist < shortestDist Then
                shortestDist = tmpDist
                nearestPoint = "endpoint"
                remPoint = points(n)
            End If

        Next

        Select Case nearestPoint
            Case "startpoint"
                pl.AddVertexAt(0, remPoint.Position.Convert2d(New Plane), 0, 0, 0)
            Case ("endpoint")
                pl.AddVertexAt(pl.NumberOfVertices, remPoint.Position.Convert2d(New Plane), 0, 0, 0)
            Case Else
                Err.Raise(0, , "nearestPoint not set!")
        End Select

        Return RemovePoint(points, remPoint)

    End Function

    Private Function GetShortestLine(ByRef points() As DBPoint) As Polyline

        Dim n As Long
        Dim m As Long

        'Info that needs to be captured
        Dim basePoint As DBPoint
        Dim endPoint As DBPoint
        Dim shortestDist As Double = 1.0E+300 'A very very big number!

        'Loop through every combination of point pairs, and find the pair with the shortest distance
        For n = 0 To points.GetUpperBound(0)
            For m = n + 1 To points.GetUpperBound(0)

                Dim tmpDist As Double
                tmpDist = points(n).Position.DistanceTo(points(m).Position)

                'If this is the shortest distance so far, update
                If tmpDist < shortestDist Then
                    shortestDist = tmpDist
                    basePoint = points(n)
                    endPoint = points(m)
                End If

            Next
        Next

        'Add the points forming the shortest distance to a new polyline
        Dim pl As Polyline = New Polyline
        pl.AddVertexAt(0, basePoint.Position.Convert2d(New Plane), 0, 0, 0)
        pl.AddVertexAt(1, endPoint.Position.Convert2d(New Plane), 0, 0, 0)

        'remove basePoint and endPoint from the array of points, so that they are not reused
        RemovePoint(points, basePoint)
        RemovePoint(points, endPoint)

        Return pl
    End Function

    Private Function RemovePoint(ByRef points() As DBPoint, ByRef remPoint As DBPoint) As Boolean
        Dim n As Long
        Dim newUpperBound As Long = points.GetUpperBound(0) - 1
        Dim pointFound As Boolean
        'Iterate through the points array until the removePoint is found, then nudge points down the array
        For n = 0 To newUpperBound
            If Not pointFound Then
                If points(n) Is remPoint Then
                    pointFound = True
                End If
            End If
            If pointFound Then
                points(n) = points(n + 1)
            End If
        Next
        'Set the new size of the array, clipping off the last item of the array
        ReDim Preserve points(newUpperBound)

        If newUpperBound = -1 Then
            Return True
        Else
            Return False
        End If

    End Function

    Private Sub AddToModelSpace(ByVal db As Database, ByVal myT As Transaction, ByVal ent As Entity)

        'Open the ModelSpace Block Table Record
        Dim acBT As BlockTable = db.BlockTableId.GetObject(OpenMode.ForRead)
        Dim BTR As BlockTableRecord
        BTR = acBT(BlockTableRecord.ModelSpace).GetObject(OpenMode.ForWrite)

        'Add the entity to the ModelSpace Block Table Record
        BTR.AppendEntity(ent)

        'Add the entity to the transaction
        myT.AddNewlyCreatedDBObject(ent, True)

    End Sub

End Class
